// File RTCG2.cs --- generating and calling a method with argument and result
// sestoft@dina.kvl.dk * 2003-03-18

using System;
using System.Reflection;
using System.Reflection.Emit;

class RTCG2 {
  public static void Main(String [] args) {
    AssemblyName assemblyName = new AssemblyName();
    assemblyName.Name = "myassembly";
    // Build: run-only assembly
    AssemblyBuilder assemblyBuilder = 
      AppDomain.CurrentDomain.DefineDynamicAssembly(assemblyName,
                                                    AssemblyBuilderAccess.Run);
    // Build: module mymodule
    ModuleBuilder moduleBuilder = 
      assemblyBuilder.DefineDynamicModule("mymodule");
    // Build: public class MyClass : IMyInterface { ... }
    TypeBuilder typeBuilder = 
      moduleBuilder.DefineType("MyClass", 
                               TypeAttributes.Class | TypeAttributes.Public, 
                               typeof(Object));
    typeBuilder.AddInterfaceImplementation(typeof(IMyInterface));
    { // Build constructor: public MyClass() : base() { }
      ConstructorBuilder constructorBuilder = 
        typeBuilder.DefineConstructor(MethodAttributes.Public,
                                      CallingConventions.HasThis,
                                      new Type[] { });
      ILGenerator ilg = constructorBuilder.GetILGenerator();
      ilg.Emit(OpCodes.Ldarg_0);      // push this
      ilg.Emit(OpCodes.Call, typeof(Object).GetConstructor(new Type[] {}));
      ilg.Emit(OpCodes.Ret);
    }
    {
      // Build method: public virtual double MyMethod(double x) { ... }
      MethodBuilder methodBuilder =
        typeBuilder.DefineMethod("MyMethod", 
                                 MethodAttributes.Virtual | MethodAttributes.Public,
                                 typeof(double),
                                 new Type[] { typeof(double) });
      // Obtain an IL generator to build the method body
      ILGenerator ilg = methodBuilder.GetILGenerator();
      ilg.EmitWriteLine("MyClass.MyMethod() was called");
      ilg.Emit(OpCodes.Ldarg_1);
      ilg.Emit(OpCodes.Ldc_R8, 2.1);
      ilg.Emit(OpCodes.Add);
      ilg.Emit(OpCodes.Ret);  
    }
    Type ty = typeBuilder.CreateType();

    // Call 1: Call MyClass.MyMethod(5) via interface call:
    Object obj = ty.GetConstructor(new Type[] {}).Invoke(new Object[] { });
    IMyInterface myObj = (IMyInterface)obj;
    double res = myObj.MyMethod(5);
    Console.WriteLine(res);

    // Call 2: Call MyClass.MyMethod(6) using reflection:
    res = (double)ty.GetMethod("MyMethod").Invoke(obj, new object[] { 6 });
    Console.WriteLine(res);

    // Call 3: Call MyClass.MyMethod(7) as delegate:
    D2D mm = (D2D)Delegate.CreateDelegate(typeof(D2D), obj, "MyMethod");
    res = mm(7);
    Console.WriteLine(res);
  }

  public delegate double D2D(double x);
}

public interface IMyInterface {
  double MyMethod(double x);
}
